package org.nst.pasientlink.authenticator;

/**
 * <p>Title: Utility</p>
 * <p>Description: misc utility functions used by the authenticator class</p>
 * <p>Copyright: Copyright (c) 2002</p>
 * <p>Company: Norwegian Centre for Telemedicine </p>
 * @author Thomas Strandens, Per Atle Bakkevoll
 * @version 1.0
 */

import java.util.*;
import java.io.IOException;
import java.io.InputStream;
import java.security.*;
import java.text.DecimalFormat;
import java.lang.Math;
import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;

public class Utility {
   private AuthenticatorStorage myStore; // persistence handler
   private SecureRandom randomizer; // object for generating single use passvoid's
   private static int singleUsePasswordLength = 6;
   protected static final String AUTHENTICATOR_CONFIGURATION_PROPERTIES = "/authenticator.properties";
   private MessageDigest digest; // hash generator
   private static Logger logger = Logger.getLogger("authenticator"); // logger used throughout the application
   private static Object SMSsemaphore = new Object(); // object used for synchronizing multiple SMS sending threads
   private static Properties configuration; // contains configuration properties

   /**
    * constructor. Loads configuration and sets up logger and secure random generator
    * (this takes a while, and it's fair to do it here instead of waiting for
    * the first request arrives)
    *
    * @param myAuthenticator Authenticator instance
    */

   public Utility() {
      loadConfiguration();

      configureLogger();
      // set up secure random generator (and generate a single random byte stream to jump-start the object)
      try {
         this.randomizer = SecureRandom.getInstance(Authenticator.SECURE_RANDOM_ALGORITHM, Authenticator.CRYPTOGRAPHY_PROVIDER);
         byte[] intializeBytes = new byte[32];
         this.randomizer.nextBytes(intializeBytes);
      }
      catch (Exception e) {
         logger.fatal("Unable to initialize secure random generator (algorithm: " + Authenticator.SECURE_RANDOM_ALGORITHM + ", provider: " + Authenticator.CRYPTOGRAPHY_PROVIDER + "): " + e.getMessage());
      }
      try {
         this.singleUsePasswordLength = new Integer(this.configuration.getProperty("single_use_password_length")).intValue();
      }
      catch(Exception e) {
         logger.error("Unable to set single use password length from configuration, defaulting to " + this.singleUsePasswordLength);
      }
   }

   /**
    * Main. Makes a BASE64-encoded SHA-1 hash from the the first input parameter,
    * and outputs this on the console. Invoked by the Ant build file in order to
    * generate administration passwords used in the configuration file.
    *
    * @param args String[] contains command line parameters
    */

   public static void main(String[] args) {
      Utility myUtil = new Utility();
      if(args.length > 0) {
         sun.misc.BASE64Encoder myEncoder = new sun.misc.BASE64Encoder();
         System.out.println("BASE64SHA-1: " + myEncoder.encode(myUtil.hashString(args[0])));
      }

   }

   /**
    * verify that the specified password hash compares successfully with the
    * stored password hash, for the specified user. we assume that the user id
    * exists in the system, i.e. that this has been verified before this method
    * is invoked (but we won't crash horribly if this verification has not been done).
    *
    * @param myUserID String containing the user id of the user password to be verified
    * @param myPasswordHash byte[] containing the password hash of the user to be verified
    * @return boolean indicating whether the two supplied passwords are equal (true) or not (false)
    */

   protected boolean verifyPasswordHash(String myUserID, byte[] myPasswordHash) {
      byte[] userPasswordHash = null;
      boolean retVal = false;
      if(myUserID != null &&
         myPasswordHash != null) {
            userPasswordHash = this.myStore.getPasswordHash(myUserID);
            if(userPasswordHash != null) {
               // compare byte arrays for length and content
               if(userPasswordHash.length == myPasswordHash.length &&
                  (byteArrayCompare(userPasswordHash, myPasswordHash) == true)) {
                  retVal = true;
               }
            }
      }
      return retVal;
   }

   /**
    * generateSingleUsePassword generates and returns single-use password consisting
    * of a number of digits. The size depends on the configuration property "single_use_password_length"
    * The method uses the secure random generator to generate the single use password.
    *
    * @return String single-use password generated using the secure random generator
    *
    * @see org.nst.pasientlink.authenticator.Authenticator
    */

   public synchronized String generateSingleUsePassword() {
      long myLongVal;

      myLongVal = this.randomizer.nextLong();
      String singleUsePassword = new Long(Math.abs(myLongVal)).toString().substring(0,this.singleUsePasswordLength);
      if(singleUsePassword.length() < this.singleUsePasswordLength) {
         char[] leftPad = new char[this.singleUsePasswordLength - singleUsePassword.length()];
         Arrays.fill(leftPad, '0');
         singleUsePassword = new String(leftPad) + singleUsePassword;
      }
      return singleUsePassword;
   }

   /**
    * sendSMSNotification spawns a thread to send a GSM SMS single use password notification.
    * This method does not wait for the SMS message to actually be send, it returns when the
    * message is queued for delivery. When the adminGSMSubscriberID field is set, no checking
    * against the database for user existence is done.
    *
    * @param myRequest Request containing single use password, userid etc. to be sent
    * @param adminGSMSubscriberID String containing admin GSM subscriber ID, or null if this is not an admin request
    * @return boolean indicating whether the message was prepared for sending successfully (true) or not (false)
    *
    */

   public boolean sendSMSNotification(Request myRequest, String adminGSMSubscriberID) {
      boolean retVal = false;
      if(myRequest != null &&
         adminGSMSubscriberID == null && // this is not an admin request...
         myRequest.getUserID() != null &&
         myRequest.getSingleUsePassword() != null) {

         String GSMSubscriberID = this.myStore.getGSMSubscriberID(myRequest.getUserID());
         if(GSMSubscriberID != null) {
            new SendSMSThread(this.SMSsemaphore, myRequest, GSMSubscriberID).start();
            retVal = true;
            // note - above thread performs logging, both info and fatal
         }
      }
      else if(myRequest != null &&
              adminGSMSubscriberID != null && // this is an admin request...
              myRequest.getUserID() != null &&
              myRequest.getSingleUsePassword() != null) {

            new SendSMSThread(this.SMSsemaphore, myRequest, adminGSMSubscriberID).start();
            retVal = true;
      }
      return retVal;
   }

   /**
    * createUser creates a user from the specified input
    * parameters, given that a user with the same userid does not exist.
    *
    * @param myUserID String containing userID of the new user
    * @param myUserDescription String containing a description, typically name, of the new user
    * @param myPassword String containing password for the new user
    * @param myGSMSubscriberID String containing the GSM phone number of the new user
    * @return a boolean value indicating whether the user was created (true) or not (false)
    *
    */

   protected boolean createUser(String myUserID,
                                String myUserDescription,
                                String myPassword,
                                String myGSMSubscriberID) {
		boolean retVal = false;

      if(myUserID != null &&
         myPassword != null &&
         myGSMSubscriberID != null) {

         if(!myStore.userExists(myUserID)) { // ensure that user does not exist
            byte[] myPasswordHash = hashString(myPassword);
            if(myStore.createUser(myUserID, myUserDescription, myPasswordHash, myGSMSubscriberID)) {
               retVal = true;
            }
         }
         else {
            logger.warn("CreateUser: UserId '" + myUserID + "' already exists.");
         }
      }
      return retVal;
   }

   /**
    * hashString returns a hash of the input String
    *
    * @param myString String to be hashed
    * @return String containing hash of input string
    */

   public synchronized byte[] hashString(String myString) {
      if(myString != null) {
         if(this.digest == null) {
            try {
               this.digest = MessageDigest.getInstance(Authenticator.PASSWORD_HASH_ALGORITHM, Authenticator.CRYPTOGRAPHY_PROVIDER);
            }
            catch(NoSuchAlgorithmException e) {
               logger.fatal("Unable to set up message hash object: " + e.getMessage());
            }
            catch(NoSuchProviderException e) {
               logger.fatal("Unable to set up message hash object: " + e.getMessage());
            }
         }
         return this.digest.digest(myString.getBytes());
      }
      else {
         return null;
      }
   }

   /**
    * byteArrayCompare compares two byte arrays. The two byte arrays should
    * be of equal size. If they are not, the comparison will stop at the last
    * byte of the smallest byte array.
    *
    * @param a byte[] first byte array
    * @param b byte[] second byte array
    * @return boolean indicating whether the byte arrays are equal (true) or not (false)
    *
    */

   public boolean byteArrayCompare(byte[] a, byte[] b) {
      boolean retVal = true;
      int len = a.length;
      if (len > b.length) len = b.length;
      for (int i = 0; i < len; ++i) {
         if (a[i] != b[i]) {
            retVal = false;
            break;
         }
      }
      return retVal;
   }

   /**
    * configureLogger configures the logger
    *
    */

   private void configureLogger() {
		// Configure log4j
		if (this.configuration != null) {
		   PropertyConfigurator.configure(this.configuration);

		} else {
			logger.fatal("could not configure logger, verify existence of properties configuration file");
		}
   }

   /**
    * getLogger returns the logger instance to be used throughout the application
    *
    * @return Logger instance
    */

   public static Logger getLogger() {
      return logger;
   }

   /**
    * getConfiguration is called by misc. classes to get the configuration properties
    * for the authenticator
    *
    * @return Properties containing configuration
    */

   public static Properties getConfiguration() {
      return configuration;
   }

   /**
    * loadConfiguration loads configuration properties from file
    * using the getResourceAsStream method.
    */

   protected void loadConfiguration() {
      InputStream myStream = null;
      try {
         myStream = getClass().getResourceAsStream(AUTHENTICATOR_CONFIGURATION_PROPERTIES);
         if(myStream != null) {
            this.configuration = new Properties();
            this.configuration.load(myStream);
            myStream.close();
         }
      }
      catch(IOException e) {
         logger.fatal("could not load properties configuration file \"" + AUTHENTICATOR_CONFIGURATION_PROPERTIES + "\".");
      }
   }


   /**
    * setAuthenticatorStorage configures the storage class used by utility
    *
    * @param myAuthenticatorStorage AuthenticatorStorage object to be used
    */

   public void setAuthenticatorStorage(AuthenticatorStorage myAuthenticatorStorage) {
      this.myStore = myAuthenticatorStorage;
   }

   /**
    * getSingleUsePasswordLength returns the size of the single use password,
    * as loaded from the configuration properties file
    *
    * @return int containing the size of the generated single use passwords
    */

   public int getSingleUsePasswordLength() {
      return this.singleUsePasswordLength;
   }
}